package Semant;

public class Semant {
    private Env env;
    private Translate.Level level;
    private Temp.Label _break; // which is the last break label?
    public Translate.Translate codegen;
    static Types.Type INT = new Types.INT();
    static Types.Type STRING = new Types.STRING();
    static Types.Type VOID = new Types.VOID();
    static Types.Type NIL = new Types.NIL();

    public Semant(ErrorMsg.ErrorMsg err) { this(new Env(err)); }

    private Semant(Env e) {
	this(e, new Translate.Level(new Intel.IntelFrame()),
	     new Translate.Translate(),null);
    }

    private Semant(Env e, Translate.Level l, Translate.Translate c,
		   Temp.Label b) {
	codegen=c; env=e; level=l; _break=b;
    }

    // The main semantic function
    public Translate.Exp transProg(Absyn.Exp exp) {
	FindEscape.FindEscape dummy = new FindEscape.FindEscape(exp);

	return transExp(exp).exp;
    }

    private Util.BoolList extractEscapes(Absyn.FieldList p) {
	if (p == null) return null;

	return new Util.BoolList(p.escape, extractEscapes(p.tail));
    }

    private Translate.Exp checkInt(ExpTy et, int pos)
    {
	if (et.ty != INT) env.errorMsg.error(pos,"integer required");

	return et.exp;
    }

    // Can t2 be assigned to t1 ?
    private void checkCompatible(Types.Type t1,Types.Type t2,int pos)
    {
	// t1 = t1.actual(); // t2 = t2.actual();

	if (t1 == t2 && !(t2 instanceof Types.NIL)) return;
	if (t2 instanceof Types.NIL && t1 instanceof Types.RECORD) return;

	env.errorMsg.error(pos,"incompatible types");
    }

    // Are et1.ty and et2.ty compatible in = or in if_then_else?
    // If v then void's and nil's are considered unifiable
    private Types.Type checkUnifiable(ExpTy et1, ExpTy et2, int pos, boolean v)
    {
	if (et1.ty == et2.ty && (et1.ty == INT || et1.ty == STRING ||
				 et1.ty instanceof Types.RECORD
				 || et1.ty instanceof Types.ARRAY))
	    return et1.ty;

	if (v && et1.ty == et2.ty && (et1.ty == VOID || et1.ty == NIL))
	    return et1.ty;

	if (et1.ty == NIL && et2.ty instanceof Types.RECORD) return et2.ty;

	if (et2.ty == NIL && et1.ty instanceof Types.RECORD) return et1.ty;

	env.errorMsg.error(pos,"incompatible types");

	return null;
    }

    // Dispatcher
    ExpTy transVar(Absyn.Var l, boolean assigned) {
	if (l instanceof Absyn.SimpleVar)
	    return transVar((Absyn.SimpleVar)l,assigned);
	if (l instanceof Absyn.SubscriptVar)
	    return transVar((Absyn.SubscriptVar)l,assigned);
	if (l instanceof Absyn.FieldVar)
	    return transVar((Absyn.FieldVar)l,assigned);

	// we should never reach this point
	throw new Error("transVar");
    }

    // cases
    ExpTy transVar(Absyn.SimpleVar v, boolean assigned) {
	Entry x = (Entry)env.venv.get(v.name);

	if (!(x instanceof VarEntry)) {
	    env.errorMsg.error(v.pos,"undefined variable " +
			       v.name.toString());
	    return new ExpTy(null,INT);
	}

	VarEntry ent = (VarEntry)x;

	if (assigned && ent.constant) {
	    env.errorMsg.error(v.pos,"variable " + v.name.toString() +
			       " cannot be assigned to");
	    return new ExpTy(null,INT);
	}

	return new ExpTy(codegen.SimpleVar(ent.access,level),ent.ty.actual());
    }

    ExpTy transVar(Absyn.SubscriptVar v, boolean assigned) {
	ExpTy array = transVar(v.var,assigned);
	ExpTy index = transExp(v.index);

	if (array.ty instanceof Types.ARRAY) {
	    checkInt(index,v.index.pos);

	    return new ExpTy(codegen.SubscriptVar(array.exp,index.exp,level),
			     ((Types.ARRAY)array.ty).element.actual());
	}
	else {
	    env.errorMsg.error(v.pos,"array type required");
	    return new ExpTy(null,INT);
	}
    }

    ExpTy transVar(Absyn.FieldVar v, boolean assigned) {
	int offset = 0;
	ExpTy var = transVar(v.var,assigned);

	if (var.ty instanceof Types.RECORD) {
	    Types.RECORD Cursor;

	    for (Cursor = ((Types.RECORD)var.ty); Cursor!=null;
		 offset++, Cursor = Cursor.tail)
		if (Cursor.fieldName == v.field) break;

	    if (Cursor == null) {
		env.errorMsg.error(v.pos,"field " + v.field.toString() +
				   "does not exist");
		return new ExpTy(null,INT);
	    };

	    return new ExpTy(codegen.FieldVar(var.exp,offset,level),
			     Cursor.fieldType.actual());
	}
	else {
	    env.errorMsg.error(v.pos,"record type required");
	    return new ExpTy(null,INT);
	}
    }

    // Dispatcher
    ExpTy transExp(Absyn.Exp e) {
	if (e instanceof Absyn.VarExp) return transExp((Absyn.VarExp)e);
	if (e instanceof Absyn.NilExp) return transExp((Absyn.NilExp)e);
	if (e instanceof Absyn.IntExp) return transExp((Absyn.IntExp)e);
	if (e instanceof Absyn.StringExp) return transExp((Absyn.StringExp)e);
	if (e instanceof Absyn.CallExp) return transExp((Absyn.CallExp)e);
	if (e instanceof Absyn.OpExp) return transExp((Absyn.OpExp)e);
	if (e instanceof Absyn.RecordExp) return transExp((Absyn.RecordExp)e);
	if (e instanceof Absyn.SeqExp) return transExp((Absyn.SeqExp)e);
	if (e instanceof Absyn.AssignExp) return transExp((Absyn.AssignExp)e);
	if (e instanceof Absyn.IfExp) return transExp((Absyn.IfExp)e);
	if (e instanceof Absyn.WhileExp) return transExp((Absyn.WhileExp)e);
	if (e instanceof Absyn.ForExp) return transExp((Absyn.ForExp)e);
	if (e instanceof Absyn.BreakExp) return transExp((Absyn.BreakExp)e);
	if (e instanceof Absyn.LetExp) return transExp((Absyn.LetExp)e);
	if (e instanceof Absyn.ArrayExp) return transExp((Absyn.ArrayExp)e);

	// We should never reach this point
	throw new Error("transExp");
    }

    // cases
    ExpTy transExp(Absyn.VarExp e) {
	ExpTy et = transVar(e.var,false);

	return new ExpTy(codegen.VarExp(et.exp),et.ty);
    }

    ExpTy transExp(Absyn.NilExp e) {
	return new ExpTy(codegen.NilExp(),NIL);
    }

    ExpTy transExp(Absyn.IntExp e) {
	return new ExpTy(codegen.IntExp(e.value),INT);
    }

    ExpTy transExp(Absyn.StringExp e) {
	return new ExpTy(codegen.StringExp(e.value,level),STRING);
    }

    ExpTy transExp(Absyn.CallExp e) {
	Entry ent = (Entry)(env.venv.get(e.func));

	if (ent == null || !(ent instanceof FunEntry)) {
	    env.errorMsg.error(e.pos,"unknown function identifier "
			       + e.func.toString());
	    return new ExpTy(null,INT);
	}

	FunEntry f = (FunEntry)ent;

	Absyn.ExpList actuals = e.args;
	Types.RECORD formals = f.formals;
	ExpTy et = null;
	Translate.ExpList args = null, cursor = null;

	for (; actuals != null; actuals = actuals.tail,
		  formals = formals.tail) {
	    if (formals == null) {
		env.errorMsg.error(e.pos,"too many arguments");
		break;
	    }

	    et = transExp(actuals.head);
	    checkCompatible(formals.fieldType.actual(),et.ty,actuals.head.pos);

	    if (args == null)
		args = cursor = new Translate.ExpList(et.exp,null);
	    else {
		cursor.tail = new Translate.ExpList(et.exp,null);
		cursor = cursor.tail;
	    }
	}

	if (formals != null) env.errorMsg.error(e.pos,"too few arguments");

	return new ExpTy(codegen.CallExp(f.label,args,f.level,level,
					 f.result.actual() == VOID),
			 f.result.actual());
    }

    ExpTy transExp(Absyn.OpExp e) {
	ExpTy et1 = transExp(e.left);
	ExpTy et2 = transExp(e.right);

	if (e.oper != Absyn.OpExp.EQ && e.oper != Absyn.OpExp.NE) {
	    checkInt(et1,e.left.pos);
	    checkInt(et2,e.right.pos);
	    return new ExpTy(codegen.OpExp(et1.exp,e.oper,et2.exp),INT);
	}
	else {
	    // EQ and NE can also be applied to RECORDs and ARRAYs
	    checkUnifiable(et1,et2,e.pos,false);

	    return new ExpTy(codegen.OpExp(et1.exp,e.oper,et2.exp),INT);
	}
    }

    ExpTy transExp(Absyn.RecordExp e) {
	Types.Type type = (Types.Type)env.tenv.get(e.typ);

	if (type == null) {
	    env.errorMsg.error(e.pos,"undefined type " + e.typ.toString());
	    return new ExpTy(null,INT);
	}

	type = type.actual();
	if (!(type instanceof Types.RECORD)) {
	    env.errorMsg.error(e.pos,"record identifier required");
	    return new ExpTy(null,INT);
	}

	Absyn.FieldExpList a = e.fields;
	Types.RECORD f = (Types.RECORD)type;
	Translate.ExpList codes = null, cursor = null;
	ExpTy et;

	int pos = e.pos;

	for (; a != null; a = a.tail, f = f.tail) {
	    pos = a.pos;

	    if (f == null) {
		env.errorMsg.error(a.pos,"field "+ a.name.toString() +
				   " is unknown");
		break;
	    }

	    if (f.fieldName != a.name) {
		env.errorMsg.error(a.pos,f.fieldName.toString() +
				   " field expected");
		break;
	    }

	    et = transExp(a.init);
	    checkCompatible(f.fieldType.actual(),et.ty,a.pos);
	    if (codes == null)
		codes = cursor = new Translate.ExpList(et.exp,null);
	    else {
		cursor.tail = new Translate.ExpList(et.exp,null);
		cursor = cursor.tail;
	    }
	}

	if (f != null) env.errorMsg.error(pos,"missing record fields");

	return new ExpTy(codegen.RecordExp(codes),type);
    }

    ExpTy transExp(Absyn.SeqExp e) {
	Types.Type type = VOID;
	Translate.Exp code = null;
	ExpTy et;

	for (Absyn.ExpList h = e.list; h != null; h = h.tail) {
	    et = transExp(h.head);
	    if (code != null)
		code = codegen.Append(code,et.exp);
	    else
		code = et.exp;

	    type = et.ty;
	}

	return new ExpTy((code == null)? codegen.Nop() : code,type);
    }

    ExpTy transExp(Absyn.AssignExp e) {
	ExpTy vt = transVar(e.var,true);
	ExpTy et = transExp(e.exp);

	checkCompatible(vt.ty,et.ty,e.pos);

	return new ExpTy(codegen.AssignExp(vt.exp,et.exp),VOID);
    }

    ExpTy transExp(Absyn.IfExp e) {
	Types.Type temp;

	ExpTy test = transExp(e.test);
	ExpTy thenclause = transExp(e.thenclause);
	ExpTy elseclause = (e.elseclause != null) ? transExp(e.elseclause)
	    : null;

	checkInt(test,e.test.pos);

	if (elseclause != null)
	    if ((temp = checkUnifiable(thenclause,elseclause,
				       e.elseclause.pos,true)) != null)
		return new ExpTy(codegen.IfExp(test.exp,thenclause.exp,
					       elseclause.exp),temp);
	    else return new ExpTy(null,VOID);

	if (thenclause.ty == VOID)
	    return new ExpTy(codegen.IfExp(test.exp,thenclause.exp),VOID);

	env.errorMsg.error(e.thenclause.pos,"then clause must return void");
	return new ExpTy(null,VOID);
    }

    ExpTy transExp(Absyn.WhileExp e) {
	ExpTy test = transExp(e.test);
	checkInt(test,e.test.pos);

	Temp.Label b = new Temp.Label();

	// The body must be evaluated in a new Semant object whose break
	// label is b
	Semant newsem = new Semant(env,level,codegen,b);
	ExpTy body = newsem.transExp(e.body);

	if (body.ty != VOID) {
	    env.errorMsg.error(e.body.pos,"while's body must return void");
	    return new ExpTy(null,VOID);
	}

	return new ExpTy(codegen.WhileExp(test.exp,body.exp,b),VOID);
    }

    ExpTy transExp(Absyn.ForExp e) {
	env.venv.beginScope();
	Translate.Exp d = transDecConstant(e.var);

	VarEntry v = (VarEntry)(env.venv.get(e.var.name));
	if (v.ty != INT) env.errorMsg.error(e.var.pos,"integer required");

	ExpTy hi = transExp(e.hi);
	checkInt(hi,e.hi.pos);

	Temp.Label b = new Temp.Label();

	// The body must be evaluated in a new Semant object whose break
	// label is b
	Semant newsem = new Semant(env,level,codegen,b);
	ExpTy body = newsem.transExp(e.body);

	if (body.ty != VOID)
	    env.errorMsg.error(e.body.pos,"for's body must return void");

	env.venv.endScope();
	return new ExpTy(codegen.ForExp(d,hi.exp,body.exp,v.access,b),VOID);
    }

    ExpTy transExp(Absyn.BreakExp e) {
	if (_break == null) {
	    env.errorMsg.error(e.pos,"illegal break position");
	    return new ExpTy(null,VOID);
	}

	return new ExpTy(codegen.BreakExp(_break),VOID);
    }

    ExpTy transExp(Absyn.LetExp e) {
	Translate.Exp code = null;

	env.venv.beginScope();
	env.tenv.beginScope();

	for (Absyn.DecList p = e.decs; p != null; p = p.tail)
	    if (code == null)
		code = transDec(p.head);
	    else
		code = codegen.Append(code,transDec(p.head));

	ExpTy et = transExp(e.body);

	env.venv.endScope();
	env.tenv.endScope();

	return new ExpTy(codegen.LetExp(code,et.exp),et.ty);
    }

    ExpTy transExp(Absyn.ArrayExp e) {
	Types.Type t = (Types.Type)env.tenv.get(e.typ);

	if (t == null) {
	    env.errorMsg.error(e.pos,"undefined array type");
	    return new ExpTy(null,INT);
	}

	t = t.actual();
	if (!(t instanceof Types.ARRAY)) {
	    env.errorMsg.error(e.pos,"incompatible type");
	    return new ExpTy(null,INT);
	}

	ExpTy et1 = transExp(e.size);
	checkInt(et1,e.size.pos);

	ExpTy et2 = transExp(e.init);
	checkCompatible(((Types.ARRAY)t).element.actual(),et2.ty,e.init.pos);

	return new ExpTy(codegen.ArrayExp(et1.exp,et2.exp,level),t);
    }

    // Dispatcher
    Translate.Exp transDec(Absyn.Dec d) {
	if (d instanceof Absyn.VarDec) return transDec((Absyn.VarDec)d);
	if (d instanceof Absyn.TypeDec) return transDec((Absyn.TypeDec)d);
	if (d instanceof Absyn.FunctionDec)
	    return transDec((Absyn.FunctionDec)d);

	// We should never reach this point
	throw(new Error("transDec"));
    }

    // cases
    Translate.Exp transDec(Absyn.VarDec d) {
	ExpTy et = transExp(d.init);
	Translate.Access local;

	if (d.typ == null) {
	    if (et.ty == NIL) {
		env.errorMsg.error(d.init.pos,
				   "untyped initializer cannot be nil");
		return null;
	    }
	    else {
		local = level.allocLocal(d.escape);
		env.venv.put(d.name,new VarEntry(local,et.ty));

		return codegen.VarDec(local,et.exp,level);
	    }
	}
	else {
	    Types.Type t = transTy(d.typ).actual();
	    checkCompatible(t,et.ty,d.init.pos);

	    local = level.allocLocal(d.escape);
	    env.venv.put(d.name, new VarEntry(local,t));

	    return codegen.VarDec(local,et.exp,level);
	}
    }

    // This is like transDec(Absyn.VarDec), but the new variable cannot
    // be assigned to
    Translate.Exp transDecConstant(Absyn.VarDec d) {
	ExpTy et = transExp(d.init);
	Translate.Access local;

	if (d.typ == null) {
	    if (et.ty == NIL) {
		env.errorMsg.error(d.init.pos,
				   "untyped initializer cannot be nil");
		return null;
	    }
	    else {
		local = level.allocLocal(d.escape);
		env.venv.put(d.name,new VarEntry(local,et.ty,true));

		return codegen.VarDec(local,et.exp,level);
	    }
	}
	else {
	    Types.Type t = transTy(d.typ).actual();
	    checkCompatible(t,et.ty,d.init.pos);

	    local = level.allocLocal(d.escape);
	    env.venv.put(d.name, new VarEntry(local,t,true));

	    return codegen.VarDec(local,et.exp,level);
	}
    }

    Translate.Exp transDec(Absyn.TypeDec d) {
	boolean flag = true;
	int Counter = 0;
	Absyn.TypeDec start = d, Temp = null;
	Translate.Exp code = null;

	// first we check that no two types in the same bunch
	// have the same name
	for (; d != null; d = d.next) {
	    for (Temp = d.next; Temp != null; Temp = Temp.next)
		if (d.name == Temp.name) {
		    env.errorMsg.error(Temp.pos,"Redefinition of " + d.name +
				       " in the same block of definitions");
		    return null;
		}

	    code = (code == null)? codegen.TypeDec() :
		codegen.Append(code,codegen.TypeDec());
	}

	for (d = start; d != null; Counter++, d = d.next)
	    env.tenv.put(d.name,new Types.NAME(d.name));

	while (flag) {
	    flag = false;
	    Counter--;
	    if (Counter < 0) {
		env.errorMsg.error(start.pos,
				   "illegal cyclic type declarations");
		return null;
	    }

	    for (d = start; d != null; d = d.next) {
		Types.Type t = transTy(d.ty);

		// We guarantee that a NAME is never bound to another NAME
		if (t instanceof Types.NAME && !(((Types.NAME)t).bound()))
		    flag = true;
		else
		    ((Types.NAME)(env.tenv.get(d.name))).bind(t.actual());
	    }
	}

	return code;
    }

    Translate.Exp transDec(Absyn.FunctionDec d) {
	Types.Type result;
	Absyn.FunctionDec Temp1 = d, Temp2 = null;
	Translate.Level newlevel;
	Translate.Exp code = null;

	// first we collect the signatures of the function declarations
	for (; d != null; d = d.next) {
	    // we check that no other function in this bunch has the same name
	    for (Temp2 = d.next; Temp2 != null; Temp2 = Temp2.next)
		if (d.name == Temp2.name) {
		    env.errorMsg.error(Temp2.pos,"Redefinition of " + d.name +
				       " in the same block of definitions");
		    return null;
		}

	    result = (d.result != null) ? transTy(d.result).actual() : VOID;

	    Types.RECORD formals = transTypeFields(d.params);

	    Temp.Label name = new Temp.Label(d.name);
	    newlevel = new Translate.Level(level,name,
					   extractEscapes(d.params));

	    env.venv.put(d.name,new FunEntry(newlevel,name,formals,result));
	}

	// then we can type-check their bodies
	for (d = Temp1; d != null; d = d.next) {
	    FunEntry f=(FunEntry)env.venv.get(d.name);

	    env.venv.beginScope();

	    Absyn.FieldList p = d.params;
	    Translate.AccessList al = f.level.formals;
	    for (; p != null; p = p.tail, al = al.tail) {
		Types.Type temp = (Types.Type)env.tenv.get(p.typ);

		if (temp == null) {
		    env.errorMsg.error(p.pos,"undeclared type " +
				       p.typ.toString());
		    env.venv.endScope();
		    return null;
		}

		env.venv.put(p.name,new VarEntry(al.head,temp));
	    }

	    // we create a new Semant objects where breaks are not allowed
	    // (until a for or while is found)
	    Semant newsem = new Semant(env,f.level,codegen,null);

	    ExpTy et = newsem.transExp(d.body);
	    env.venv.endScope();

	    checkCompatible(f.result.actual(),et.ty,d.body.pos);

	    code = (code == null) ?
		codegen.FunctionDec(f.label,et.exp,newsem.level) :
		codegen.Append(code,codegen.FunctionDec
			       (f.label,et.exp,newsem.level));
	}

	return code;
    }

    private Types.RECORD transTypeFields(Absyn.FieldList p)
    {
	Types.RECORD result=null, cursor=null;

	for (; p!=null; p=p.tail) {
	    Types.Type t = ((Types.Type)(env.tenv.get(p.typ))).actual();
	    if (t == null) {
		env.errorMsg.error(p.pos,"undefined type " + p.typ.toString());
		continue;
	    }

	    if (cursor == null)
		result = cursor = new Types.RECORD(p.name,t,null);
	    else {
		cursor.tail = new Types.RECORD(p.name,t,null);
		cursor = cursor.tail;
	    }
	}

	return result;
    }

    // Dispatcher
    Types.Type transTy(Absyn.Ty t) {
	if (t instanceof Absyn.NameTy) return transTy((Absyn.NameTy)t);
	if (t instanceof Absyn.RecordTy) return transTy((Absyn.RecordTy)t);
	if (t instanceof Absyn.ArrayTy) return transTy((Absyn.ArrayTy)t);

	// We should never reach this point
	throw new Error("transTy");
    }

    Types.Type transTy(Absyn.NameTy t) {
	Types.Type Result = (Types.Type)(env.tenv.get(t.name));

	if (Result != null) return Result;

	env.errorMsg.error(t.pos,"undefined type " + t.name.toString());
	return INT;
    }

    Types.Type transTy(Absyn.RecordTy t) {
	Types.RECORD Result = null, Cursor = null;
	Types.Type temp;

	for (Absyn.FieldList fs = t.fields; fs != null; fs = fs.tail)
	    {
		temp = (Types.Type)(env.tenv.get(fs.typ));
		if (temp == null) {
		    env.errorMsg.error(fs.pos,"undefined type " +
				       fs.typ.toString());
		    return INT;
		}

		if (Cursor != null) {
		    Cursor.tail = new Types.RECORD(fs.name,temp,null);
		    Cursor = Cursor.tail;
		}
		else
		    Cursor = Result = new Types.RECORD(fs.name,temp,null);
	    };

	return Result;
    }

    Types.Type transTy(Absyn.ArrayTy t) {
	Types.Type temp = (Types.Type)(env.tenv.get(t.typ));

	if (temp == null) {
	    env.errorMsg.error(t.pos,"undefined type " + t.typ.toString());
	    return INT;
	}

	return new Types.ARRAY(temp);
    }
}
